#include "precompile.h"

#include <glm/glm.hpp>
#include <glm/gtc/matrix_transform.hpp>
#include <glm/vec2.hpp>

#include "buff/select_target_with_spec_pos.h"

#include "creature.h"

#include "mt/Buff.h"

void SelectTargetWithSpecPosBuff::Activate()
{
    hold_param3_ = meta->GetBuffParam3(this);
    glm::vec3 center = glm::vec3
        (
         meta->GetBuffParam5(this),
         meta->GetBuffParam6(this),
         meta->GetBuffParam7(this)
         );
    std::vector<Creature*> targets;
    bool include_car = false;
    bool exclude_passenger = false;
    if (!meta->_buff_param8_int_set.empty()) {
        include_car = meta->_buff_param8_int_set.find(1) != meta->_buff_param8_int_set.end();
        exclude_passenger = meta->_buff_param8_int_set.find(2) != meta->_buff_param8_int_set.end();
    }
    owner->TraverseCreatures
        (
         [this, &targets, &center, include_car, exclude_passenger] (Creature* c, bool& stop)
         {
             if (c->dead ||
                 (c->IsCar() && !include_car) ||
                 a8::HasBitFlag(c->status, CS_Disable) ||
                 c->HasBuffEffect(kBET_Dive)
                 ) {
                 return;
             }
             if (exclude_passenger) {
                 if (c->HasBuffEffect(kBET_Driver) &&
                     c->HasBuffEffect(kBET_Passenger)) {
                     return;
                 }
             }
             float distance = c->GetPos().Distance2D2(center);
#ifdef MYDEBUG
             if (c->IsPlayer()) {
                             a8::XPrintf("select spec pos distance:%f center:%f %f %f pos:%f %f %f range:%f\n",
                                         {
                                             distance,
                                                 center.x,
                                                 center.y,
                                                 center.z,
                                                 c->GetPos().GetX(),
                                                 c->GetPos().GetY(),
                                                 c->GetPos().GetZ(),
                                                 hold_param3_,
                                         });
             }
#endif
             if (distance < hold_param3_) {
                 switch (meta->_int_buff_param1) {
                 case kBST_All:
                     {
                         targets.push_back(c);
                     }
                     break;
                 case kBST_Self:
                     {
                         if (c == owner) {
                             targets.push_back(c);
                         }
                     }
                     break;
                 case kBST_FriendlyIncludeSelf:
                     {
                         if (c->team_id == owner->team_id) {
#ifdef MYDEBUG
                             a8::XPrintf("select spec pos distance:%f center:%f %f %f pos:%f %f %f range:%f\n",
                                         {
                                             distance,
                                                 center.x,
                                                 center.y,
                                                 center.z,
                                                 c->GetPos().GetX(),
                                                 c->GetPos().GetY(),
                                                 c->GetPos().GetZ(),
                                                 hold_param3_,
                                         });
#endif
                             targets.push_back(c);
                         }
                     }
                     break;
                 case kBST_FriendlyExcludeSelf:
                     {
                         if (c->team_id == owner->team_id && c != owner) {
                             targets.push_back(c);
                         }
                     }
                     break;
                 case kBST_Enemy:
                     {
                         if (c->team_id != owner->team_id) {
                             targets.push_back(c);
                         }
                     }
                     break;
                 case kBST_EnemyAndSelf:
                     {
                         if (c->team_id != owner->team_id || c == owner) {
                             targets.push_back(c);
                         }
                     }
                     break;
                 default:
                     {
                     }
                     break;
                 }
             }
         }
         );
    for (auto& target : targets) {
        for (int buff_id : meta->_buff_param2_int_list) {
            target->TryAddBuff(caster_.Get(), buff_id, skill_meta, init_args, buff_vars);
        }
    }
}

void SelectTargetWithSpecPosBuff::Deactivate()
{

}
