#include "precompile.h"

#include "custom_battle.h"
#include "custom_team.h"
#include "custom_member.h"
#include "netdata.h"
#include "roommgr.h"
#include "room.h"
#include "player.h"

#include "cs_proto.pb.h"
#include "ss_proto.pb.h"

#include "mt/Map.h"
#include "mt/MapMode.h"
#include "mt/Param.h"

void CustomBattle::Init()
{
    ob_team_ = std::make_shared<CustomTeam>();
    ob_team_->team_uuid_ = "";
    ob_team_->is_view_ = true;
    f8::Timer::Instance()->SetIntervalEx
        (
         2000,
         [this] (int event, const a8::Args* args)
         {
             if (a8::TIMER_EXEC_EVENT == event) {
                 NotifyState();
             }
         },
         &timer_attacher);
}

void CustomBattle::UnInit()
{

}

void CustomBattle::ParseResult(a8::XObject& obj)
{
    if (!obj.HasKey("errcode")) {
        parse_ok_ = false;
        return;
    }
    room_id_ = obj.Get("room_id").GetString();
    room_uuid_ = obj.Get("room_uuid").GetString();
    zone_id_ = obj.Get("zone_id");
    node_id_ = obj.Get("node_id");
    map_mode_ = obj.Get("mode_id");
    map_id_ = obj.Get("map_id");
    match_mode_ = obj.Get("match_mode");
    start_time_ = obj.Get("start_time");
    sign_ = obj.Get("sign").GetString();
    const mt::Map* map_meta = mt::Map::GetById(map_id_);
    if (!map_meta || !map_meta->IsOpen()) {
        parse_ok_ = false;
        return;
    }
    map_mode_meta_ = mt::MapMode::GetById(map_mode_);
    if (!map_mode_meta_) {
        parse_ok_ = false;
        return;
    }
    {
        is_moba_ = map_meta->is_moba() ? true : false;
    }
    {
        auto team_list = obj.At("team_list");
        if (!team_list || !team_list->IsArray()) {
            parse_ok_ = false;
            return;
        }
        for (int i = 0;i < team_list->Size(); ++i) {
            auto team_obj = team_list->At(i);
            if (!team_obj || !team_obj->IsObject()) {
                parse_ok_ = false;
                return;
            }
            std::string team_uuid = team_obj->Get("team_uuid").GetString();
            int is_view = team_obj->Get("is_view").GetInt();
            auto member_list = team_obj->At("members");
            if (!member_list ||
                !member_list->IsArray()) {
                parse_ok_ = false;
                return;
            }
            auto team = GetTeamByTeamUuid(team_uuid);
            if (!team) {
                team = std::make_shared<CustomTeam>();
                team->custom_battle_ = this;
                team->team_uuid_ = team_uuid;
                team->is_view_ = false;
                uuid_hash_[team->team_uuid_] = team;
                team_list_.push_back(team);
            }
            for (int ii = 0; ii < member_list->Size(); ++ii) {
                auto member_obj = member_list->At(ii);
                if (!member_obj || !member_obj->IsObject()) {
                    parse_ok_ = false;
                    return;
                }
                auto member = std::make_shared<CustomMember>();
                member->team_ = team.get();
                member->account_id_ = member_obj->Get("account_id").GetString();
                member->session_id_ = member_obj->Get("session_id").GetString();
                member->name_ = member_obj->Get("name").GetString();
                member->sex_ = member_obj->Get("sex");
                member->robot_id_ = member_obj->Get("robot_id");
                member->is_android_ = member_obj->Get("is_android");
                member->avatar_url_ = member_obj->Get("head_id").GetString();
                member->battle_context_ = std::make_shared<BattleDataContext>();
                member->battle_context_->ParseResult(*member_obj);
                if (!member->battle_context_->parse_ok) {
                    parse_ok_ = false;
                    return;
                }
                member_id_hash_[member->account_id_] = member;
                team->member_hash_[member->account_id_] = member;
                account_hash_[member->account_id_] = team;
            }
        }
    }
    {
        auto ob_list = obj.At("ob_list");
        if (ob_list && ob_list->IsArray()) {
            for (int i = 0; i < ob_list->Size(); ++i) {
                auto member_obj = ob_list->At(i);
                if (!member_obj || !member_obj->IsObject()) {
                    parse_ok_ = false;
                    return;
                }
                auto member = std::make_shared<CustomMember>();
                member->team_ = ob_team_.get();
                member->account_id_ = member_obj->Get("account_id").GetString();
                member->session_id_ = member_obj->Get("session_id").GetString();
                member->name_ = member_obj->Get("name").GetString();
                member->sex_ = member_obj->Get("sex");
                member->battle_context_ = std::make_shared<BattleDataContext>();
                member->battle_context_->ParseResult(*member_obj);
                if (!member->battle_context_->parse_ok) {
                    parse_ok_ = false;
                    return;
                }
                account_hash_[member->account_id_] = ob_team_;
                ob_id_hash_[member->account_id_] = member;
            }
        }
    }

    TraverseTeamList
        (
         [this] (std::shared_ptr<CustomTeam> team) -> bool
         {
             team->TraverseMember
                 (
                  [this] (std::shared_ptr<CustomMember> m) -> bool
                  {
                      auto member = m;
                      f8::Timer::Instance()->SetTimeoutEx
                          (
                           a8::RandEx(mt::Param::s().battle_auto_ready_min_time,
                                      mt::Param::s().battle_auto_ready_max_time) * 1000,
                           [member] (int event, const a8::Args* args)
                           {
                               if (a8::TIMER_EXEC_EVENT == event) {
                                   if (!member->is_ready_) {
                                       member->is_ready_ = 1;
                                       member->GetTeam()->GetCustomBattle()->OnMemberReady(member.get());
                                   }
                               }
                           },
                           &timer_attacher);
                      return true;
                  });
             return true;
         });

    raw_data_ = std::make_shared<a8::XObject>();
    obj.DeepCopy(*raw_data_);
    CalcTeam1AverageHeroLv();
    parse_ok_ = true;
}

std::shared_ptr<CustomMember> CustomBattle::GetMemberByAccountId(const std::string& account_id)
{
    auto itr = member_id_hash_.find(account_id);
    return itr != member_id_hash_.end() ? itr->second : nullptr;
}

std::shared_ptr<CustomTeam> CustomBattle::GetTeamByAccountId(const std::string& account_id)
{
    auto itr = account_hash_.find(account_id);
    return itr != account_hash_.end() ? itr->second : nullptr;
}

std::shared_ptr<CustomTeam> CustomBattle::GetTeamByTeamUuid(const std::string& team_uuid)
{
    auto itr = uuid_hash_.find(team_uuid);
    return itr != uuid_hash_.end() ? itr->second : nullptr;
}

bool CustomBattle::AllIsJoined()
{
    bool ok = true;
    for (auto& pair : member_id_hash_) {
        if (!pair.second->IsJoined()) {
            ok = false;
            break;
        }
    }
    return ok;
}

bool CustomBattle::AllIsReady()
{
    bool ok = true;
    for (auto& pair : member_id_hash_) {
        if (!pair.second->IsReady()) {
            ok = false;
            break;
        }
    }
    return ok;
}

int CustomBattle::GetMemberNum()
{
    return member_id_hash_.size();
}

int CustomBattle::GetTeamNum()
{
    return uuid_hash_.size();
}

void CustomBattle::TraverseMemberList(std::function<bool (CustomMember*)> func)
{
    for (auto& pair : member_id_hash_) {
        if (!func(pair.second.get())) {
            break;
        }
    }
}

void CustomBattle::TraverseTeamList(std::function<bool (std::shared_ptr<CustomTeam>)> cb)
{
    for (auto& pair : uuid_hash_) {
        if (!cb(pair.second)) {
            break;
        }
    }
}

std::shared_ptr<CustomMember> CustomBattle::GetObByAccountId(const std::string& account_id)
{
    auto itr = ob_id_hash_.find(account_id);
    return itr != ob_id_hash_.end() ? itr->second : nullptr;
}

void CustomBattle::TraverseObList(std::function<bool (std::shared_ptr<CustomMember>)> cb)
{
    for (auto& pair : ob_id_hash_) {
        if (!cb(pair.second)) {
            break;
        }
    }
}

RoomMode_e CustomBattle::GetRoomMode()
{
    if (IsMoba()) {
        return kMobaMode;
    } else {
        if (match_mode_) {
            if (GetMapModeMeta()->mapMode() == mt::kPvpRankMode) {
                return kPvpRankMode;
            } else {
                return kPvpMode;
            }
        } else {
            return kPvpMode;
        }
    }
}

void CustomBattle::CalcTeam1AverageHeroLv()
{
    if (!team_list_.empty()) {
        team1_average_hero_lv_ = team_list_.at(0)->GetAverageHeroLv();
    }
}

RoomType_e CustomBattle::GetRoomType()
{
    if (IsMoba()) {
        return RoomType_MidBrid;
    }
    if (IsCustomMode()) {
        return RoomType_MidBrid;
    }
    if (GetRoomMode() == kPvpRankMode) {
        auto rank_mode_conf = mt::Param::GetRankModeConfByHeroLv(team1_average_hero_lv_);
        if (!rank_mode_conf) {
            abort();
        }
        return rank_mode_conf->room_type;
    } else {
        if (team1_average_hero_lv_ < mt::Param::s().new_room_max_level) {
            return RoomType_OldBrid1;
        }
        if (team1_average_hero_lv_ < mt::Param::s().mid_room_max_level) {
            return RoomType_OldBrid2;
        }
        return RoomType_OldBrid3;
    }
}

std::shared_ptr<CustomTeam> CustomBattle::GetTeamByIdx(int idx)
{
    if (GetTeamNum() <= 0) {
        return nullptr;
    }
    if (idx < 0 || idx >= GetTeamNum()) {
        return nullptr;
    }
    return team_list_.at(idx);
}

void CustomBattle::OnEnter(std::shared_ptr<cs::CMJoin> join_msg, long ip_saddr, int socket_handle,
                           std::weak_ptr<SocketDisconnectHandler> sd_handler)
{
    if (sd_handler.expired()) {
        socket_handle = 0;
    }
    auto member = GetMemberByAccountId(join_msg->account_id());
    if (!member) {
        member = GetObByAccountId(join_msg->account_id());
        if (!member) {
            return;
        }
    }
    {
        ss::SS_BindUpStreamSocket notifymsg;
        GGListener::Instance()->SendToClient(socket_handle, 0, notifymsg);
    }
    NotifyState();
    RoomMgr::Instance()->UnBindCustomMemberSocket(socket_handle);
    member->join_msg_ = join_msg;
    member->ip_saddr_ = ip_saddr;
    member->socket_handle_ = socket_handle;
    if (socket_handle) {
        socket_hash_[socket_handle] = member;
        RoomMgr::Instance()->ReBindCustomMemberSocket(member);
    }
    NotifyState();
    if (IsPvp()) {
        for (auto& pair : member_id_hash_) {
            pair.second->is_ready_ = 1;
            OnMemberReady(pair.second.get());
        }
    }
#ifdef MYDEBUG
    a8::XPrintf("OnEnter socket_handle:%d\n", {socket_handle});
#endif
    if (state_ == 1) {
        auto hum = GetRoom()->GetPlayerByAccountId(join_msg->account_id());
        if (hum) {
            hum->ReJoin(ip_saddr, socket_handle);
        }
    }
}

void CustomBattle::OnMemberReady(CustomMember* member)
{
    if (state_ == 0 && AllIsReady()) {
        state_ = 1;
        GameStart();
    }
    NotifyState();
}

void CustomBattle::GameStart()
{
    if (IsNormalMode()) {
        auto room = RoomMgr::Instance()->GetJoinableRoom(shared_from_this());
        if (!room) {
            RoomType_e self_room_type = GetRoomType();
            if (IsPvp()) {
                room = RoomMgr::Instance()->CreateRoom
                    (self_room_type,
                     GetMapModeMeta()->id(),
                     GetMapId(),
                     GetRoomMode(),
                     nullptr);
            } else {
                room = RoomMgr::Instance()->CreateRoom
                    (self_room_type,
                     GetMapModeMeta()->id(),
                     GetMapId(),
                     GetRoomMode(),
                     nullptr);
            }
        }
        SetRoom(room.get());
        room->JoinWithCustomBattle(shared_from_this());
    } else {
        abort();
    }
}
