#include "precompile.h"

#include "mt/BattleHeroGrow.h"
#include "mt/Buff.h"

IMPL_TABLE(mt::BattleHeroGrow)

std::map<long long, const mt::BattleHeroGrow*> mt::BattleHeroGrow::level_hash_;

namespace mt
{

    void BattleHeroGrow::Init1()
    {
        attrs_ = std::make_shared<std::vector<std::tuple<int, float>>>();
        effects_ = std::make_shared<std::vector<int>>();
        if (GetByHeroIdLevel(heroId(), level())) {
            abort();
        }
        level_hash_[a8::MakeInt64(heroId(), level())] = this;
        {
            std::vector<std::string> strings;
            a8::Split(currentLevelAttribute(), strings, '|');
            for (auto& str : strings) {
                std::vector<std::string> strings2;
                a8::Split(str, strings2, ':');
                if (strings2.size() != 2) {
                    abort();
                }
                int attr_id = a8::XValue(strings2.at(0));
                float attr_val = a8::XValue(strings2.at(1)).GetDouble();
                if (!IsValidHumanAttr(attr_id) &&
                    !IsValidHumanVirtualAttr(attr_id)) {
                    abort();
                }
                attrs_->push_back(std::make_tuple(attr_id, attr_val));
            }
        }
        {
            std::vector<std::string> strings;
            a8::Split(currentEquipEffect(), strings, '|');
            for (auto& str : strings) {
                int buff_id = a8::XValue(str);
                if (!mt::Buff::GetById(buff_id)) {
                    abort();
                }
                effects_->push_back(buff_id);
            }
        }
    }

    void BattleHeroGrow::Init2()
    {
        next_meta_ = GetByHeroIdLevel(heroId(), level() + 1);
    }

    const BattleHeroGrow* BattleHeroGrow::GetNextMeta() const
    {
        return next_meta_;
    }

    bool BattleHeroGrow::IsFullLevel() const
    {
        return levelExp() <= 0 || !GetNextMeta();
    }

    std::shared_ptr<std::vector<std::tuple<int, float>>> BattleHeroGrow::GetLevelAttrs() const
    {
        return attrs_;
    }

    std::shared_ptr<std::vector<int>> BattleHeroGrow::GetEquipEffects() const
    {
        return effects_;
    }

    int BattleHeroGrow::GetSkillEffect() const
    {
        return currentSkillEffect();
    }

    const BattleHeroGrow* BattleHeroGrow::GetByHeroIdLevel(int hero_id, int level)
    {
        auto itr = level_hash_.find(a8::MakeInt64(hero_id, level));
        return itr != level_hash_.end() ? itr->second : nullptr;
    }

}
