#pragma once

#include "mt/macro.h"
#include "mtb/Map.h"

namespace mt
{
    class MapCollider;

    struct WorldObject
    {
        int object_uniid = 0;
        int object_id = 0;
        int object_type = 0;
        glm::vec3 pos;
        glm::vec3 size;
        float rotation = 0.0f;
    };

    DECLARE_ID_TABLE(Map, mtb::Map,
                     "map@map.json",
                     "map_id")
    public:
        std::vector<std::tuple<std::string, int>> template_list_;
        int rand_space = 0;
        std::vector<int> airdrops_;
        std::vector<int> airraids_;
        int refresh_robot_min_num = 0;
        int refresh_robot_max_num = 0;
        int refresh_robot_min_time = 0;
        int refresh_robot_max_time = 0;
        std::vector<int> buff_list;
        glm::vec3 first_safearea_center_;
        std::map<int, int> car_num_limit_;
        std::vector<int> safearea_list;
        std::vector<std::shared_ptr<WorldObject>> _world_objects;
        std::map<int, std::vector<std::shared_ptr<WorldObject>>> _group_world_objects;
        std::shared_ptr<MapCollider> collider_info = nullptr;
        std::vector<std::tuple<std::shared_ptr<WorldObject>, int>> moba_born_points;
        std::vector<std::vector<std::tuple<std::shared_ptr<WorldObject>, int>>> moba_path_points;

        std::string RandTemplate() const;
        int GetCarLimit(int car_id) const;
        glm::vec3 GetCenter() const;
        int RandSafeArea() const;
        bool IsPveMap() const;
        bool IsOpen() const;
        glm::vec3 GroundSamplingPos() const;
        const MobaRoom* GetMobaRoomMeta() const { return moba_room_meta_; }

        void Init1();
        void Init2();

    private:
        glm::vec3 sampling_pos_;
        const MobaRoom* moba_room_meta_ = nullptr;

        void PostProcess();
    };

}
