#include "precompile.h"

#include "mt/MapArea.h"

IMPL_TABLE(mt::MapArea)
std::map<int, std::vector<const mt::MapArea*>> mt::MapArea::map_areas;

namespace mt
{
    void MapArea::Init1()
    {
        auto itr = map_areas.find(map_id());
        if (itr != map_areas.end()) {
            itr->second.push_back(this);
        } else {
            map_areas[map_id()] = std::vector<const mt::MapArea*>({this});
        }
        {
            std::vector<std::string> strings;
            a8::Split(area_center(), strings, ':');
            if (strings.size() != 2) {
                abort();
            }
            center_ = glm::vec2(a8::XValue(strings[0]).GetDouble(), a8::XValue(strings[1]).GetDouble());
        }
    }

    const mt::MapArea* MapArea::GetAreaByPoint(int map_id, float x, float y, float z)
    {
        auto itr = map_areas.find(map_id);
        if (itr != map_areas.end()) {
            for (auto area_meta : itr->second) {
                if (x < area_meta->center_.x - area_meta->area_width() / 2) {
                    continue;
                }
                if (x > area_meta->center_.x + area_meta->area_width() / 2) {
                    continue;
                }
                if (z < area_meta->center_.y - area_meta->area_height() / 2) {
                    continue;
                }
                if (z > area_meta->center_.y + area_meta->area_height() / 2) {
                    continue;
                }
                return area_meta;
            }
        }
        return nullptr;
    }

}
