#include "precompile.h"

#include "base_agent.h"
#include "tracemgr.h"
#include "perf.h"

#include <f8/btmgr.h>


BaseAgent::BaseAgent():behaviac::Agent()
{
}

BaseAgent::~BaseAgent()
{
}

int BaseAgent::GetTickCount()
{
    return 0;
}

int BaseAgent::RandRange(int min_val, int max_val)
{
    return a8::RandEx(min_val, max_val);
}

int BaseAgent::Rand()
{
    return rand();
}

float BaseAgent::RandRangeAsFloat(float min_val, float max_val)
{
    return RandRange(min_val, max_val);
}

void BaseAgent::Exec()
{
    behaviac::EBTStatus status = f8::BtMgr::Instance()->BtExec(this);
    if (status == behaviac::BT_RUNNING) {
        CheckCoroutineEvent();
    }
}

behaviac::EBTStatus BaseAgent::StartCoroutine(std::shared_ptr<BtCoroutine> co)
{
    coroutines_hash_[co->GetId()] = co;
    list_add_tail(&coroutines_list_, &co->entry);
    return behaviac::BT_RUNNING;
}

bool BaseAgent::PreEnterCoroutine(int co_id, behaviac::EBTStatus& status)
{
    status = behaviac::BT_FAILURE;
    if (coroutines_hash_.empty()) {
        return false;
    }
    auto itr = coroutines_hash_.find(co_id);
    if (itr == coroutines_hash_.end()) {
        return false;
    }
    auto& co = itr->second;
    if (co->status != behaviac::BT_RUNNING) {
        abort();
    }
    if (co->IsAbort()) {
        list_del_init(&co->entry);
        coroutines_hash_.erase(itr);
        return false;
    }
    co->status = co->runing_cb(co.get());
    status = co->status;
    if (co->status != behaviac::BT_RUNNING) {
        list_del_init(&co->entry);
        coroutines_hash_.erase(itr);
    }
    return true;
}

void BaseAgent::CheckCoroutineEvent()
{
    int co_id = -1;
    bool has_abort_co = false;
    {
        list_head* pos = nullptr;
        list_head* next = nullptr;
        list_head* head = &coroutines_list_;
        list_for_each_safe(pos, next, head) {
            BtCoroutine* co = list_entry(pos, BtCoroutine, entry);
            if (co->GetContext()->HasEvent()) {
                co_id = co->GetId();
                break;
            }
        }
    }
    if (co_id >= 0) {
        auto itr = coroutines_hash_.find(co_id);
        if (itr == coroutines_hash_.end()) {
            abort();
        }
        auto co = itr->second;
        coroutines_hash_.clear();
        INIT_LIST_HEAD(&coroutines_list_);
        co->GetContext()->FireEvent(this);
    }
}

int BaseAgent::DeltaTime(int time)
{
    return GetTickCount() - time;
}

void BaseAgent::AbortCoroutine(int co_id)
{
    auto itr = coroutines_hash_.find(co_id);
    if (itr != coroutines_hash_.end()) {
        if (!itr->second->IsAbort()) {
            itr->second->Abort(0);
            itr->second->runing_cb(itr->second.get());
            itr->second->runing_cb = nullptr;
            list_del_init(&itr->second->entry);
            coroutines_hash_.erase(itr);
        }
    }
}
