#pragma once

#include <f8/protoutils.h>

#include "cs_proto.pb.h"

namespace a8
{
    class WebSocketClient;
}

class AndroidAgent;
class Player : public std::enable_shared_from_this<Player>
{
 public:
    enum { HID = HID_Player };

public:

    void Update();

    void Init(int idx, const std::string& account_id);
    void UnInit();
    template <typename T>
    void SendMsg(T& msg)
    {
        static int msgid = f8::Net_GetMessageId(msg);
        InternalSendMsg(msgid, msg);
    }
    int GetSocketId() { return socket_handle_; }
    bool NetConnected() { return net_connected_; }
    const std::string& GetAccountId() { return account_id_; }
    const std::string& GetSessionId() { return session_id_; }
    void SetSessionId(const std::string& session_id) { session_id_ = session_id; }
    void SetHeroUniId(long long hero_uniid) { hero_uniid_ = hero_uniid; }
    long long GetHeroUniId() { return hero_uniid_; }
    auto GetWebSocket() { return web_socket_; }
    int GetIdx() { return idx_; }
    void SetLoginSucess();
    void SetLoginFail();
    bool IsLoginSucess() { return is_login_success_; };
    bool IsJoinOk() { return join_ok_; }

    void _SMKcpHandshake(f8::MsgHdr* hdr, const cs::SMKcpHandshake& msg);
    void _SMPing(f8::MsgHdr* hdr, const cs::SMPing& msg);
    void _SMRpcError(f8::MsgHdr* hdr, const cs::SMRpcError& msg);
    void _SMReconnect(f8::MsgHdr* hdr, const cs::SMReconnect& msg);
    void _SMWatchWar(f8::MsgHdr* hdr, const cs::SMWatchWar& msg);
    void _SMLeave(f8::MsgHdr* hdr, const cs::SMLeave& msg);
    void _SMJoinedNotify(f8::MsgHdr* hdr, const cs::SMJoinedNotify& msg);
    void _SMMapInfo(f8::MsgHdr* hdr, const cs::SMMapInfo& msg);
    void _SMUpdate(f8::MsgHdr* hdr, const cs::SMUpdate& msg);
    void _SMRollMsg(f8::MsgHdr* hdr, const cs::SMRollMsg& msg);
    void _SMVoiceNotify(f8::MsgHdr* hdr, const cs::SMVoiceNotify& msg);
    void _SMDisconnectNotify(f8::MsgHdr* hdr, const cs::SMDisconnectNotify& msg);
    void _SMGameOver(f8::MsgHdr* hdr, const cs::SMGameOver& msg);
    void _SMDebugMsg(f8::MsgHdr* hdr, const cs::SMDebugMsg& msg);
    void _SMUiUpdate(f8::MsgHdr* hdr, const cs::SMUiUpdate& msg);
    void _SMGameStart(f8::MsgHdr* hdr, const cs::SMGameStart& msg);
    void _SMSysPiaoMsg(f8::MsgHdr* hdr, const cs::SMSysPiaoMsg& msg);
    void _SMShowCountdown(f8::MsgHdr* hdr, const cs::SMShowCountdown& msg);
    void _SMShowTeamUI(f8::MsgHdr* hdr, const cs::SMShowTeamUI& msg);
    void _SMGetItemNotify(f8::MsgHdr* hdr, const cs::SMGetItemNotify& msg);
    void _SMPvePassWave(f8::MsgHdr* hdr, const cs::SMPvePassWave& msg);
    void _SMTeamMarkTargetPosList(f8::MsgHdr* hdr, const cs::SMTeamMarkTargetPosList& msg);
    void _SMDebugCmd(f8::MsgHdr* hdr, const cs::SMDebugCmd& msg);
    void _SMNewBieEnd(f8::MsgHdr* hdr, const cs::SMNewBieEnd& msg);

private:
    void InternalSendMsg(int msgid, ::google::protobuf::Message& msg);

private:
    AndroidAgent* agent_ = nullptr;

    int idx_ = 0;
    int socket_handle_ = 0;
    std::string account_id_;
    std::string session_id_;
    std::string remote_ip_;
    int remote_port_ = 0;
    long long hero_uniid_ = 0;

    bool is_login_success_ = false;
    bool net_connected_ = false;
    bool join_ok_ = false;
    long long last_update_tick_ = 0;

    std::shared_ptr<a8::WebSocketClient> web_socket_;

};
